"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.QueryRequest = exports.QueryBuilder = void 0;
const mongodb_1 = require("mongodb");
const mongodb_2 = require("../connections/mongodb");
const errorBuilder_1 = require("../errorBuilder");
const validateValueClass_1 = require("../helpers/validateValueClass");
const metadata_1 = require("../global/metadata");
const relationship_1 = require("../decorators/interfaces/relationship");
const getQueries_1 = require("../helpers/getQueries");
const pickData_1 = require("../helpers/pickData");
const type_1 = require("../helpers/type");
const bcryptjs_1 = __importDefault(require("bcryptjs"));
const commons_1 = require("../helpers/commons");
class QueryBuilder {
    constructor() {
        this.withRelations = [];
    }
    getTableName() {
        return ((0, commons_1.convertToPlural)(this.constructor.name));
    }
    getColumns() {
        return (0, metadata_1.getMetadataTable)().getColumn(this.constructor);
    }
    getNameColumns() {
        return this.getColumns().map((column) => column.name);
    }
    getRelations() {
        return (0, metadata_1.getMetadataTable)().getRelation(this.constructor);
    }
    getCollection() {
        const nameCollection = this.getTableName();
        if (mongodb_2.collections[nameCollection]) {
            return mongodb_2.collections[nameCollection];
        }
        throw Error(`collection ${nameCollection} notfound`);
    }
    with(params) {
        if (Array.isArray(params)) {
            params.forEach(e => {
                const findRelation = this.getRelations().find((relation) => relation.fieldName == e.model);
                if (findRelation) {
                    this.withRelations.push(Object.assign(Object.assign({}, e), { relation: findRelation }));
                }
                else {
                    throw errorBuilder_1.ErrorBuilder.validator(`with ${e.model.toString()} invalid`);
                }
            });
        }
        else {
            const findRelation = this.getRelations().find((relation) => relation.fieldName == params.model);
            if (findRelation) {
                this.withRelations.push(Object.assign(Object.assign({}, params), { relation: findRelation }));
            }
            else {
                throw errorBuilder_1.ErrorBuilder.validator(`with ${params.model.toString()} invalid`);
            }
        }
        return this;
    }
    findById(_id) {
        return __awaiter(this, void 0, void 0, function* () {
            if (typeof _id == "string") {
                _id = (0, type_1.objectId)(_id);
            }
            const data = yield this.findOne({ _id: { $eq: _id } });
            return data;
        });
    }
    findOne(filter) {
        return __awaiter(this, void 0, void 0, function* () {
            const data = yield this.find(filter);
            return data.length ? data[0] : null;
        });
    }
    find(filter, options) {
        return __awaiter(this, void 0, void 0, function* () {
            let data = yield this.getCollection().find(filter, options).toArray();
            if (!data)
                return [];
            const nameColumns = this.getColumns().filter(e => {
                return ((options === null || options === void 0 ? void 0 : options.fromCRUD) && !e.isPassword) || !(options === null || options === void 0 ? void 0 : options.fromCRUD);
            }).map(e => e.name);
            const dataClean = data.map((item) => (0, pickData_1.pickData)(item, nameColumns));
            const result = yield this.mappingDataRelation(dataClean, options);
            this.withRelations = [];
            return result;
        });
    }
    paginate(filter, paginate, options) {
        return __awaiter(this, void 0, void 0, function* () {
            const [data, totalCount] = yield Promise.all([
                this.find(filter, Object.assign(Object.assign({}, options), { limit: paginate.limit, skip: (paginate.page - 1) * paginate.limit })),
                this.count(filter),
            ]);
            return { data, totalCount };
        });
    }
    count(filter) {
        return __awaiter(this, void 0, void 0, function* () {
            return this.getCollection().count(filter);
        });
    }
    createOne(item) {
        return __awaiter(this, void 0, void 0, function* () {
            const newDate = new Date().getTime();
            item = yield this.validateParamBeforeSaveDB(item);
            const dataInsert = yield this.getCollection().insertOne(Object.assign(Object.assign({}, item), { createdAt: newDate, updatedAt: newDate }));
            return this.findById(dataInsert.insertedId);
        });
    }
    updateMany(filter, item) {
        return __awaiter(this, void 0, void 0, function* () {
            const newDate = new Date().getTime();
            item = yield this.validateParamBeforeSaveDB(item, true);
            const [dataFilter, _dataUpdate] = yield Promise.all([
                this.find(filter),
                this.getCollection().updateMany(filter, {
                    $set: Object.assign(Object.assign({}, item), { updatedAt: newDate }),
                })
            ]);
            return this.find({ _id: { $in: dataFilter.map(e => e._id) } });
        });
    }
    delete(filter) {
        return __awaiter(this, void 0, void 0, function* () {
            return this.getCollection().deleteOne(filter);
        });
    }
    deleteMany(filter) {
        return __awaiter(this, void 0, void 0, function* () {
            return this.getCollection().deleteMany(filter);
        });
    }
    validateParamBeforeSaveDB(params, isUpdate = false) {
        return __awaiter(this, void 0, void 0, function* () {
            (0, validateValueClass_1.validationColumnErrors)(this.constructor, params, "", isUpdate);
            yield Promise.all(this.getColumns().map((column) => __awaiter(this, void 0, void 0, function* () {
                const key = column.name;
                if (column.type == mongodb_1.ObjectId && params[key] && typeof params[key] == "string") {
                    params[key] = (0, type_1.objectId)(params[key]);
                }
                if (column.type == String && params[key] && column.isPassword) {
                    params[key] = yield bcryptjs_1.default.hash(params[key], 10);
                }
                if (column.default !== undefined && params[key] === undefined && !isUpdate) {
                    params[key] = column.default;
                }
                return;
            })));
            return params;
        });
    }
    mappingDataRelation(data, options) {
        return __awaiter(this, void 0, void 0, function* () {
            const relationMap = yield this.prepareDataRelation(data, options);
            data = data.map((item) => {
                Object.keys(relationMap).forEach((key, index) => {
                    const relation = this.getRelations()[index];
                    switch (relation.type) {
                        case relationship_1.TypeRelation.BELONG_TO:
                            item[key] = relationMap[key].find(e => {
                                let parent = e[relation.refs.PK];
                                let child = item[relation.refs.FK];
                                if (typeof child === "object")
                                    child = child.toString();
                                if (typeof parent === "object")
                                    parent = parent.toString();
                                return child == parent;
                            }) || null;
                            break;
                        case relationship_1.TypeRelation.HAS_ONE:
                            item[key] = relationMap[key].find(e => {
                                let parent = e[relation.refs.FK];
                                let child = item[relation.refs.PK];
                                if (typeof child === "object")
                                    child = child.toString();
                                if (typeof parent === "object")
                                    parent = parent.toString();
                                return child == parent;
                            }) || null;
                            break;
                        case relationship_1.TypeRelation.HAS_MANY:
                            item[key] = relationMap[key].length ? relationMap[key].filter(e => {
                                let parent = e[relation.refs.FK];
                                let child = item[relation.refs.PK];
                                if (typeof child === "object")
                                    child = child.toString();
                                if (typeof parent === "object")
                                    parent = parent.toString();
                                return child == parent;
                            }) : [];
                            break;
                        default:
                            break;
                    }
                });
                return item;
            });
            return data;
        });
    }
    prepareDataRelation(data, options) {
        return __awaiter(this, void 0, void 0, function* () {
            const optionRelations = (options === null || options === void 0 ? void 0 : options.fromCRUD) ? { fromCRUD: true } : {};
            let relationMap = yield Promise.all(this.withRelations.map((item) => {
                let ids = [];
                let dataRelation;
                const relation = item.relation;
                switch (relation.type) {
                    case relationship_1.TypeRelation.BELONG_TO:
                        ids = data.map((e) => e[relation.refs.FK]);
                        const queryBuilderParent = (0, getQueries_1.getQueryModel)(relation.refs.modelParent());
                        if (item.with) {
                            queryBuilderParent.with(item.with);
                        }
                        dataRelation = queryBuilderParent.find({
                            $and: [
                                {
                                    [relation.refs.PK]: { $in: ids },
                                },
                                Object.assign({}, item.filter),
                            ],
                        }, optionRelations);
                        break;
                    case relationship_1.TypeRelation.HAS_ONE:
                    case relationship_1.TypeRelation.HAS_MANY:
                        ids = data.map((e) => e[relation.refs.PK]);
                        const queryBuilderChild = (0, getQueries_1.getQueryModel)(relation.refs.modelChild());
                        const optionFind = relation.type == relationship_1.TypeRelation.HAS_ONE
                            ? Object.assign(Object.assign({}, optionRelations), { limit: 1 }) : Object.assign({}, optionRelations);
                        if (item.with) {
                            queryBuilderChild.with(item.with);
                        }
                        dataRelation = queryBuilderChild.find({
                            $and: [
                                {
                                    [relation.refs.FK]: { $in: ids },
                                },
                                Object.assign({}, item.filter),
                            ],
                        }, optionFind);
                        break;
                    default:
                        break;
                }
                return dataRelation;
            }));
            relationMap = relationMap.reduce((target, item, index) => {
                const relation = this.withRelations[index];
                const keyRelation = relation.alias || relation.model;
                target[keyRelation] = item;
                return target;
            }, {});
            return relationMap;
        });
    }
}
exports.QueryBuilder = QueryBuilder;
class QueryRequest {
    constructor(queryBuilder) {
        this.queryBuilder = queryBuilder;
    }
    convertQueriesIncludeObjectId(key, queryParse, resultQueries) {
        const queryColumn = queryParse[key];
        if (Array.isArray(queryColumn)) {
            resultQueries[key] = {
                $in: queryColumn.map((value) => {
                    if (typeof value != "string") {
                        throw errorBuilder_1.ErrorBuilder.validator(`Key or type ${key} not allow`);
                    }
                    return (0, type_1.objectId)(value);
                }),
            };
        }
        else {
            throw errorBuilder_1.ErrorBuilder.validator(`Key or type ${key} not allow`);
        }
        return resultQueries;
    }
    convertQueriesIncludeString(key, queryParse, resultQueries) {
        const queryColumn = queryParse[key];
        if (typeof queryColumn == "string") {
            resultQueries[key] = queryColumn;
        }
        else if (typeof queryColumn == "object") {
            const { value, isAbsolute } = queryColumn;
            if (typeof value != "string" || (isAbsolute !== undefined && typeof isAbsolute != "boolean")) {
                throw errorBuilder_1.ErrorBuilder.validator(`Key or type ${key} not allow`);
            }
            if (isAbsolute) {
                resultQueries[key] = value;
            }
            else {
                resultQueries[key] = { $regex: value, $options: "i" };
            }
        }
        else {
            throw errorBuilder_1.ErrorBuilder.validator(`Key or type ${key} not allow`);
        }
        return resultQueries;
    }
    convertQueriesIncludeNumber(key, queryParse, resultQueries) {
        const queryColumn = queryParse[key];
        if (Array.isArray(queryColumn)) {
            resultQueries[key] = {
                $in: queryColumn,
            };
        }
        else if (typeof queryColumn == "object") {
            const { isMaxExclusive, isMinExclusive, max, min } = queryColumn;
            const tmpQueriesNumber = {};
            if (max !== undefined) {
                isMaxExclusive
                    ? Object.assign(tmpQueriesNumber, { $lt: max })
                    : Object.assign(tmpQueriesNumber, { $lte: max });
            }
            if (min !== undefined) {
                isMinExclusive
                    ? Object.assign(tmpQueriesNumber, { $gt: min })
                    : Object.assign(tmpQueriesNumber, { $gte: min });
            }
            resultQueries[key] = tmpQueriesNumber;
        }
        else {
            throw errorBuilder_1.ErrorBuilder.validator(`Key or type ${key} not allow`);
        }
        return resultQueries;
    }
    /*================================+
       |CHECK VALIDATOR QUERY REQUEST   |
       +================================*/
    getQueries(queries) {
        try {
            let resultQueries = {};
            if (!queries)
                return {};
            const queryParse = JSON.parse(queries);
            if (!queryParse) {
                throw new Error("Condition fail");
            }
            const nameOfColumns = this.queryBuilder.getNameColumns();
            const columns = this.queryBuilder.getColumns();
            const checkKeyExist = Object.keys(queryParse).filter((key) => {
                return !nameOfColumns.includes(key);
            });
            if (checkKeyExist.length) {
                throw errorBuilder_1.ErrorBuilder.validator(`Key or type ${checkKeyExist} not allow`);
            }
            Object.keys(queryParse).forEach((key) => {
                const column = columns.find((x) => x.name == key);
                if (column) {
                    switch (column.type) {
                        case mongodb_1.ObjectId:
                            resultQueries = this.convertQueriesIncludeObjectId(key, queryParse, resultQueries);
                            break;
                        case String:
                            resultQueries = this.convertQueriesIncludeString(key, queryParse, resultQueries);
                            break;
                        case Number:
                            resultQueries = this.convertQueriesIncludeNumber(key, queryParse, resultQueries);
                        default:
                            break;
                    }
                }
                else {
                    throw errorBuilder_1.ErrorBuilder.validator(`Key or type ${key} not allow`);
                }
            });
            return resultQueries;
        }
        catch (error) {
            throw error;
        }
    }
    validateAndConverOperatorFilter(queries) {
        Object.keys(queries).forEach(keyObj => {
            if (keyObj == '$and' || keyObj == '$or') {
                return queries[keyObj].map(obj => this.validateAndConverOperatorFilter(obj));
            }
            else {
                const operator = Object.keys(queries[keyObj])[0];
                const field = keyObj;
                const value = queries[field][operator];
                const columns = this.queryBuilder.getColumns();
                const column = columns.find((x) => x.name == field);
                if (!column) {
                    throw errorBuilder_1.ErrorBuilder.validator(`Key or type ${field} not allow`);
                }
                if (!((operator === "$eq" || operator === "$ne") && value === null)) {
                    switch (column.type) {
                        case mongodb_1.ObjectId:
                            try {
                                if (operator == '$in' || operator == '$nin') {
                                    if (!Array.isArray(value)) {
                                        throw errorBuilder_1.ErrorBuilder.validator(`Key or type ${field} not allow`);
                                    }
                                    value.forEach(e => {
                                        if (typeof e != 'string') {
                                            throw errorBuilder_1.ErrorBuilder.validator(`Key or type ${field} not allow`);
                                        }
                                    });
                                    queries[keyObj][operator] = value.map(e => (0, type_1.objectId)(e));
                                }
                                else {
                                    queries[keyObj][operator] = (0, type_1.objectId)(value);
                                }
                            }
                            catch (error) {
                                throw errorBuilder_1.ErrorBuilder.validator(`Key or type ${field} not allow`);
                            }
                            break;
                        case String:
                            if (operator == '$in' || operator == '$nin') {
                                if (!Array.isArray(value)) {
                                    throw errorBuilder_1.ErrorBuilder.validator(`Key or type ${field} not allow`);
                                }
                                value.forEach(e => {
                                    if (typeof e != 'string') {
                                        throw errorBuilder_1.ErrorBuilder.validator(`Key or type ${field} not allow`);
                                    }
                                });
                            }
                            else {
                                if (typeof value !== 'string') {
                                    throw errorBuilder_1.ErrorBuilder.validator(`Key or type ${field} not allow`);
                                }
                            }
                            break;
                        case Number:
                            if (operator == '$in' || operator == '$nin') {
                                if (!Array.isArray(value)) {
                                    throw errorBuilder_1.ErrorBuilder.validator(`Key or type ${field} not allow`);
                                }
                                value.forEach(e => {
                                    if (typeof e !== 'number') {
                                        throw errorBuilder_1.ErrorBuilder.validator(`Key or type ${field} not allow`);
                                    }
                                });
                            }
                            else {
                                if (typeof value !== 'number') {
                                    throw errorBuilder_1.ErrorBuilder.validator(`Key or type ${field} not allow`);
                                }
                            }
                            break;
                        case Boolean:
                            if (typeof value !== 'boolean') {
                                throw errorBuilder_1.ErrorBuilder.validator(`Key or type ${field} not allow`);
                            }
                        default:
                            break;
                    }
                }
            }
        });
        return queries;
    }
}
exports.QueryRequest = QueryRequest;
